# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import sys
import os
from typing import (
    Optional,
    Mapping,
    List,
    Any
)
import logging
import logging.config
import json
from asgi_correlation_id import correlation_id
log = logging.getLogger(__name__)
#pylint: disable-next=too-many-positional-arguments
def warnings_to_logger(message: str, category: str, filename: str, lineno: str, file=None, line=None):
    #pylint: disable=unused-argument
    """replaces warnings.showwarning

    Args:
        message (str):
        category (str): Warnings class
        filename (str):
        lineno (str):
    """
    for module_name, module in sys.modules.items():
        module_path = getattr(module, "__file__", None)
        if module_path and os.path.samefile(module_path, filename):
            break
    else:
        module_name = os.path.splitext(os.path.split(filename)[1])[0]
    logger = logging.getLogger(module_name)
    extra = {"warnings.category": category, "warnings.filename": filename, "warnings.lineno": lineno}
    logger.warning(message,extra=extra)


def _make_dict_serializable(item: Any):
    secret_words = ["password", "key", "secret"]
    match item:
        case dict():
            new_dict = {}
            for k,v in item.items():
                # If value is None remove item
                if v is None:
                    continue
                key = k if isinstance(k, str) else repr(k)
                # Could also use SecretStr but changes usage of Settings object
                if any(keyword in key.lower() for keyword in secret_words):
                    new_dict[key] = "****"
                else:
                    new_dict[key] =  _make_dict_serializable(v)
            return new_dict
        case str() | int() | float():
            return item
        case list() | set():
            return [_make_dict_serializable(i) for i in item]
        case _:
            return repr(item)

class JsonFormatter(logging.Formatter):
    """Custom formatter for structured logging"""
    key_blacklist = [
        'msg','message',
        'args',
        'created', 'msecs', 'relativeCreated',
        'levelno',
        'filename',
        'color_message'
        ]
    def __init__(self,
                 datefmt: Optional[str] = "%Y-%m-%dT%H:%M:%S%z",
                 reduced: Optional[List[str]] = None,
                 indent: Optional[str] = None,
                 *, defaults: Optional[Mapping[str, Any]] = None) -> None:
        """Create a new Formatter

        Args:
            datefmt (str, optional): Used in @timestamp. Defaults to "%Y-%m-%dT%H:%M:%S%z".
            reduced (Optional[List[str]], optional): List of loglevels to reduce output by. Defaults to None.
                Reduced output removes filename, thread and process information
            indent (Optional[str], optional): indent for json dumps. Defaults to None.
        """
        super().__init__(None, datefmt, defaults=defaults)
        self.indent = indent
        self.reduced_levels = [logging.getLevelNamesMapping().get(level) for level in reduced or []]

    def _get_output_dict(self, record: logging.LogRecord) -> dict[str, Any]:
        data = {k: v for k, v in record.__dict__.items() if k not in self.key_blacklist and v is not None}
        logger_name = f"{data.pop('module')}.{data.pop('name')}"
        func_name = data.pop('funcName')
        if func_name != '<module>':
            logger_name = logger_name +"."+func_name
        output = {
            'level': data.pop("levelname"),
            'message': record.getMessage(),
            'logger_name': logger_name,
            'file': f"{data.pop('pathname')}:{data.pop('lineno')}",
            'extra': {},
            '@timestamp': self.formatTime(record,self.datefmt),
            'process': f"{data.pop('processName')}({data.pop('process')})",
            'thread': f"{data.pop('threadName')}({data.pop('thread')})",
        }
        if all((key in data for key in  ["warnings.category","warnings.filename","warnings.lineno"])):
            output['file'] = f"{data.pop('warnings.filename')}:{data.pop('warnings.lineno')}"
        if data:
            output['extra'] = _make_dict_serializable(data)
        cor_id = correlation_id.get()
        if cor_id is not None:
            output['correlationId'] = cor_id
        if self.reduced_levels and record.levelno in self.reduced_levels:
            del output['process']
            del output['logger_name']
            del output['thread']
        if not output['extra']:
            del output['extra']
        return output

    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        output = self._get_output_dict(record)
        return json.dumps(output, default=repr, indent=self.indent)


def get_logging_dict_config(level) -> dict[str,str]:
    """Generate a logging.config.dictConfig compatible dict

    Returns:
        dict: logging.config.dictConfig
    """
    default_formatter = {
        'json_formatter': {
            '()': "wurzel.utils.logging.JsonFormatter",
            'reduced': ["INFO"]
        }
    }
    default_handler= {
        "default": {
            "level" : level,
            "class": "logging.StreamHandler",
            'stream': 'ext://sys.stderr',
            "formatter": "json_formatter"
        },
    }
    logger_template = {
        "level": level,
        "handlers": ['default'],
        "propagate": False
    }
    return {
        "version": 1,
        "root": {},
        "disable_existing_loggers": False,
        "formatters": {
            **default_formatter
        },
        "handlers": {
            **default_handler
        },
        "loggers": {
            "root": {**logger_template},
            "": {**logger_template},
            "uvicorn.error": {**logger_template},
            "uvicorn.access":{
                "level": "WARNING",
                "handlers": ['default'],
                "propagate": False
            },
            'gunicorn.access': {
                "propagate": True
            },
            'gunicorn.error': {
                "propagate": True
            },
            "transaction": {**logger_template},
        },
    }



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Any, Optional, Self
from logging import getLogger

from prometheus_client import (
    REGISTRY,
    Counter, Histogram,
    push_to_gateway)

from .base_executor import (
    BaseStepExecutor,
    TypedStep,
    PathToFolderWithBaseModels,
    StepReport
)
from .settings import PrometheusSettings as Settings
log = getLogger(__name__)


class PrometheusStepExecutor(BaseStepExecutor):
    """Executer for KP steps.
    Adds PrometheusCounter; is a singelton.

    For more info see `BaseStepExecutor`.
    """
    #pylint: disable=too-many-instance-attributes
    counter_started: Counter
    counter_failed: Counter
    counter_results: Counter
    counter_inputs: Counter
    histogram_load: Histogram
    histogram_save: Histogram
    histogram_execute: Histogram
    # step_info: Info
    #Singelton Instance
    _instance = None
    s = Settings
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(PrometheusStepExecutor, cls).__new__(cls)
            cls._instance.__init__(*args, **kwargs)
            #pylint: disable=protected-access
            cls._instance.s = Settings()
            cls._instance.__setup_metrics()
        return cls._instance
    #pylint: disable=unused-private-member
    def __setup_metrics(self):
        if self.s.PROMETHEUS_DISABLE_CREATED_METRIC:
            os.environ["PROMETHEUS_DISABLE_CREATED_SERIES"] = "True"
        self.counter_started = Counter(
            "steps_started",
            "Total number of TypedSteps started",
            ("step_name",))
        self.counter_failed = Counter(
            "steps_failed",
            "Total number of TypedSteps failed",
            ("step_name",))
        self.counter_results = Counter(
            "step_results",
            "count of result, if its an array, otherwise -1",
            ("step_name",))
        self.counter_inputs = Counter(
            "step_inputs",
            "count of inputs",
            ("step_name",))
        self.histogram_save = Histogram(
            "step_hist_save",
            "Time to save results",
            ("step_name",))
        self.histogram_load = Histogram(
            "step_hist_load",
            "Time to load inputs",
            ("step_name",))
        self.histogram_execute = Histogram(
            "step_hist_execute",
            "Time to execute results",
            ("step_name",))
    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_details):
        args = {
            'gateway': self.s.PROMETHEUS_GATEWAY,
            'job': self.s.PROMETHEUS_JOB
        }
        log.info("Pushing metrics", extra=args)
        try:
            push_to_gateway(**args, registry=REGISTRY)
        except Exception:  # pylint: disable=broad-exception-caught
            log.warning(
                "Could not push prometheus metrics to gateway",
                exc_info=True)

    def execute_step(
            self,
            step_cls: TypedStep,
            inputs: Optional[set[PathToFolderWithBaseModels]],
            output_dir: Optional[PathToFolderWithBaseModels]) -> list[tuple[Any, StepReport]]:
        """Run a given Step

        Args:
            step_cls (Type[TypedDVCStep]): Step to run
            inputs (set[PathToBaseModel]): Step inputs
            output (PathToBaseModel): Step output
        """
        lbl = step_cls.__name__
        self.counter_started.labels(lbl).inc()
        try:
            data = super().execute_step(step_cls, inputs, output_dir)
        except:
            self.counter_failed.labels(lbl)
            raise
        tt_s, tt_l,tt_e = (0,0,0)
        for _, report in data:
            self.counter_results.labels(lbl).inc(report.results)
            self.counter_inputs.labels(lbl).inc(report.inputs)
            tt_s += report.time_to_save
            tt_l += report.time_to_load
            tt_e += report.time_to_execute
        self.histogram_save.labels(lbl).observe(tt_s)
        self.histogram_load.labels(lbl).observe(tt_l)
        self.histogram_execute.labels(lbl).observe(tt_e)
        return data



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import inspect
from typing import TypeAlias
from typing import ( Any,Union,get_origin)

import json
from pydantic import (
    model_validator,
    create_model
)
from pydantic_core import (
    PydanticUndefined,
    ValidationError,
    InitErrorDetails
)
from pydantic_settings import (
    SettingsConfigDict,
    BaseSettings as _PydanticBaseSettings,
    PydanticBaseSettingsSource
)


class SettingsBase(_PydanticBaseSettings):
    """Used as the PydanticSettingsBase
    Sets delimiter etc.

    When creating a custom Settings object keep these things in mind:
    1. The base settings object should inherit from SettingsBase
    2. So far only 1 layer of nesting is supported
    3. All other member settings objects need to inherit from SettingsLeaf
    4. When defining the base settings object call no SettingsLeaf constructor
    ```
        class _DBSettings(SettingsLeaf):
            USER: str
            PASSWORD: str = Field("", repr=False)
        class MySettings(SettingsBase)
            DB: _DBSettings
    ```
    """
    # Dev Note: to support n layer nesting SettingsBase
    #           needs to support the with_prefix method
    # pydantic-internal
    model_config = SettingsConfigDict(
        env_nested_delimiter='__',
        extra='forbid',
        case_sensitive=True,
        frozen=False,
        revalidate_instances='always')
    @classmethod
    #pylint: disable-next=too-many-positional-arguments
    def settings_customise_sources(cls,
        settings_cls: type[_PydanticBaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
        ) -> tuple[PydanticBaseSettingsSource, ...]:
        return init_settings, env_settings
    @model_validator(mode='before')
    @classmethod
    def before(cls, data: Any): # pylint: disable=too-many-branches
        """Before Validator, handles updating"""
        if isinstance(data, str):
            data = json.loads(data)
        if isinstance(data, dict):
            for key, field in cls.model_fields.items():
                if key in data:
                    if get_origin(field.annotation) == dict:
                        continue
                    if isinstance(data[key], str) and inspect.isclass(field.annotation) and\
                        issubclass(field.annotation, SettingsBase):
                        data[key] = json.loads(data[key])
                     # if annotation is optional and not a SettingsBase
                    elif hasattr(field.annotation,"__origin__") and \
                          field.annotation.__origin__ == Union and data[key] and \
                            inspect.isclass(field.annotation.__args__[0])\
                            and not issubclass(field.annotation.__args__[0],SettingsBase):
                        data[key] = field.annotation.__args__[0](data[key])
                        continue
                    continue
                # Is not a typing.<Dict | List | anything> and inherits from _SettingsBase
                if get_origin(field.annotation) is None:
                    if issubclass(field.annotation, SettingsLeaf):
                        try:
                            data[key] = field.annotation.with_prefix(f"{key}__")()
                        except ValidationError as verr:
                            raise ValidationError.from_exception_data(
                                verr.title,
                                [InitErrorDetails(
                                    type=err['type'],
                                    loc=(key, *err['loc']),
                                    input=err['input'],
                                    ctx={'error': err}) for err in verr.errors()],
                            )
                    elif issubclass(field.annotation, SettingsLeaf):
                        data[key] = field.annotation()
                elif field.default != PydanticUndefined:
                    data[key] = field.default
                elif field.default_factory is not None:
                    data[key] = field.default_factory()
                elif hasattr(cls,key):
                    data[key] = getattr(cls, key)
        return data
class SettingsLeaf(SettingsBase):
    """Base Class for all Setting which are not the main Setting.
    This class is mainly used to fix pydantic_settings incorrect loading of env vars.
    It might be made obsolete by future pydantic_settings releases
    https://github.com/pydantic/pydantic-settings/pull/261
    """
    @classmethod
    def with_prefix(cls, prefix: str) -> 'SettingsLeaf':
        """Returns a new class with env_prefix set"""
        cpy = create_model(prefix+"."+cls.__class__.__name__, __base__=cls)
        cpy.model_config['env_prefix'] = prefix
        return cpy
class StepSettings(SettingsLeaf):
    """Settings for Typed Steps
    In general if a class var is called `YOUR_AD_COULD_BE_HERE` the framework will try to load,
    from os.environ, a variable called `YOUR_AD_COULD_BE_HERE`.

    However for nested Settings, there will be prefixes applied separated with __ : `PARENT_VAR__YOUR_AD_COULD_BE_HERE`
    will be read from env.

    Same goes for TypedStepSettings. Their prefix is *always* `STEP_NAME_IN_UPPER_CASE`.
    ## Example
    ```python
    class KafkaSettings(Settings):
        \"\"\"Settings for Kafka\"\"\"
        # No default is given --> there must be a corresponding env_value
        BROKER_URLS: list[str] = Field([
            "kafka.example.com:9093",
            "kafka.example.com:9094",
            "kafka.example.com:9095"
            ],
            description="List of bootstrap servers")
        SASL_USER: str = "<USER_NAME>"
        SASL_PASSWORD: str
        # variables with password, key or secret won't usually be printed to terminals.
        SSL_CA_FILE: str = Field("./kafka_ca.pem", description="Path to ca file")
        TOPIC: str = "cc_mm_coupon_meta_bot"
        CONSUMER_TIMEOUT: int = Field(5*1000, description="timeout for kafka consumer in ms")
    ``` 
    """
#pylint: disable-next=invalid-name
NoSettings: TypeAlias = None
__all__=['StepSettings']



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import inspect
import importlib
import pkgutil

from typing import TypeVar, Type
from wurzel.step.typed_step import TypedStep
T = TypeVar("T", bound=object)

def find_sub_classes(parent: T, package: str=__package__) -> dict[str, T]:
    """searches for all DVC step definitions and returns them based on their name."""
    def is_non_abs_child(member: object) -> bool:
        return True and \
            inspect.isclass(member) and \
            issubclass(member, parent) and \
            not inspect.isabstract(member) and \
            not bool(getattr(member, "__abstractmethods__", False))
    result = {}
    visited = set([f'{__package__}.main', f'{__package__}.utils'])
    module_iterator = pkgutil.iter_modules(importlib.import_module(package).__path__)
    for _, module_name, is_package in module_iterator:
        full_module_name = f"{package}.{module_name}"
        if full_module_name in visited:
            continue
        visited.add(full_module_name)
        # Recurse through any sub-packages
        if is_package:
            classes_in_subpackage = find_sub_classes(parent,package=full_module_name)
            result.update(classes_in_subpackage)
        # Load the module for inspection
        module = importlib.import_module(full_module_name)

        # Iterate through all the objects in the module and
        # using the lambda, filter for class objects and only objects that exist within the module
        for _name, obj in inspect.getmembers(module, is_non_abs_child):
            result[obj.__name__] = obj
    return result

def find_typed_steps_in_package(package: str) -> dict[str, Type[TypedStep]]:
    """Recursively find all subclasses of TypedStep"""
    return find_sub_classes(TypedStep, package)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""CLI program"""
import importlib.metadata
from pathlib import Path
from typing import Type

from wurzel.step import TypedStep
from wurzel.step_executor import BaseStepExecutor
__all__ = ["generate_cli_call"]
try:
    __version__ = importlib.metadata.version('wurzel')
#pylint: disable-next=bare-except
except:
    __version__ = "dev"
__version_info__ = __version__.split(".")

def generate_cli_call(
        step_cls: Type[TypedStep],
        inputs: list[Path],
        output: Path,
        executor: Type[BaseStepExecutor]=None,
        encapsulate_env: bool = True) -> str:
    """Generate the cli call to execute a given step with its
    inputs and output

    Args:
        step_cls (Type[TypedStep]): Step to execute
        inputs (list[Path]): list of Directories
        output (Path): Output Directory

    Returns:
        str: cmd
    """
    if inputs:
        inputs = "-i " + " -i ".join((i.as_posix() for i in inputs))
    else:
        inputs = ""
    return " ".join([
        "python3 -m wurzel run",
        f"{step_cls.__module__}:{step_cls.__qualname__}",
        "-o", output.as_posix(),
        "" if executor is None else f"-e {executor.__qualname__}",
        inputs,
        "--encapsulate-env" if encapsulate_env else "--no-encapsulate-env"
    ])



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Type, Union
from types import NoneType

from pydantic import (
    create_model as py_create_model,
    BaseModel
)

from wurzel.step.typed_step import TypedStep
from wurzel.step.settings import StepSettings, SettingsBase
#pylint: disable-next=invalid-name
def WZ(typ: Type[TypedStep]):
    """Creates a Pipeline Element"""
    return TypedStep.__new__(typ)

def create_model(fields: Union[list[Union[TypedStep, Type[TypedStep]]], TypedStep], allow_extra_fields = False) -> SettingsBase:
    """Takes all fields.setting_class and creates a pydantic_settings Model.
    - If input is a list:
        - If list of Type[TypedStep]: WZ(item) will be applied
        - If list of TypedStep (WZ was already aplpied): nothing
    - if input single TypedStep: required_steps will be traversed
        - Minimum step itself



    Model takes the form of:
    ```python
        # For each step in fields:
        class MetaSettings_UPPERCASESTEPNAME(SettingsLeaf):
            # for each field_name, value pair in step.settings_cls
            field_name: field_annotation = value
            ...
        ...
        class MetaSettings_parent(SettingsBase):
            UPPERCASESTEPNAME: MetaSettings_UPPERCASESTEPNAME
            ...
    ```

    Args:
        fields (list[Union[TypedStep, Type[TypedStep]]]): will be fields

    Returns:
        SettingsBase: MetaModel with set fields
    """
    def clean(flds: Union[list[Union[TypedStep, Type[TypedStep]]], TypedStep]) -> list[TypedStep]:
        if isinstance(flds, TypedStep):
            return list(flds.traverse())
        cleaned: list[TypedStep] = [WZ(f) if isinstance(f, type) else f for f in flds]
        return cleaned
    clean_fields = clean(fields)
    base_class = py_create_model("SettingsLeaf_allow_extra", __base__=StepSettings)
    if allow_extra_fields:
        base_class.model_config['extra'] = 'allow'
    inner_models: dict[str, StepSettings] = {
        step.__class__.__name__.upper(): py_create_model(
            'MetaSettings_'+ step.__class__.__name__,
            **{name: (v.annotation,v) for name,v in step.settings_class.model_fields.items()},
            __base__=base_class
        ) for step in clean_fields if step.settings_class != NoneType}


    new_model_class: Type[BaseModel] = py_create_model(
        'MetaSettings_Parent',
        **{name: (typ, ...) for name, typ in inner_models.items()},
        __base__=SettingsBase
    )
    return new_model_class



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""CLI Entry"""
import os
import importlib
import inspect
from typing import Annotated
from pathlib import Path
import pkgutil
from datetime import datetime
import logging
import logging.config
import typer
import typer.core

from wurzel.utils.logging import get_logging_dict_config
from wurzel import TypedStep
from wurzel.steps import __all__ as all_steps
from wurzel.step_executor import (
    BaseStepExecutor,
    PrometheusStepExecutor
)

from wurzel.cli.cmd_run import main as cmd_run
from wurzel.cli.cmd_inspect import main as cmd_inspect
app = typer.Typer(
    no_args_is_help=True,
)
log = logging.getLogger(__name__)
packages = [p for p in pkgutil.iter_modules() if p.ispkg and p.name.startswith(("steps", "wurzel"))]
def executer_callback(_ctx: typer.Context, _param: typer.CallbackParam, value: str):
    """Convert a cli-str to a Type[BaseStepExecutor]

    Args:
        _ctx (typer.Context)
        _param (typer.CallbackParam):
        value (str): user typed string

    Raises:
        typer.BadParameter: If user typed string does not correlate with a Executor

    Returns:
        Type[BaseStepExecutor]: {BaseStepExecutor, PrometheusStepExecutor}
    """
    if 'BASESTEPEXECUTOR'.startswith(value.upper()):
        return BaseStepExecutor
    if 'PROMETHEUSSTEPEXECUTOR'.startswith(value.upper()):
        return PrometheusStepExecutor
    raise typer.BadParameter(f"{value} is not a recognized executor")

def step_callback(_ctx: typer.Context, _param: typer.CallbackParam, import_path: str):
    """Converts a cli-str to a TypedStep

    Args:
        _ctx (typer.Context): 
        _param (typer.CallbackParam): 
        path (str): user-typed string

    Raises:
        typer.BadParameter: import not possible

    Returns:
        Type[TypedStep]: <<step>>
    """
    try:
        if ':' in import_path:
            mod, kls = import_path.rsplit(':', 1)
        else:
            mod, kls = import_path.rsplit('.', 1)
        module = importlib.import_module(mod)
        step = getattr(module, kls)
        assert inspect.isclass(step) and issubclass(step, TypedStep)
    except ValueError as ve:
        raise typer.BadParameter(
            "Path is not in correct format, should be module.submodule.Step"
        ) from ve
    except ModuleNotFoundError as me:
        raise typer.BadParameter(
            f"Module '{mod}' could not be imported"
        ) from me
    except AttributeError as ae:
        raise typer.BadParameter(
            f"Class '{kls}' not in module {module}"
        ) from ae
    except AssertionError as ae:
        raise typer.BadParameter(
            f"Class '{kls}' not a TypedStep"
        ) from ae
    return step

def complete_step_import(_incomplete: str):
    """AutoComplete for steps
    Currently only supports library steps"""
    hints = [
        f"wurzel.steps.{step}"
        for step in all_steps if step.endswith("Step") and step != 'TypedStep'
    ]
    return hints
@app.command(
    no_args_is_help=True,
    help="Run a step"
)
#pylint: disable-next=dangerous-default-value
def run(
    step: Annotated[str, typer.Argument(
        allow_dash=False,
        help="module path to step",
        autocompletion=complete_step_import,
        callback=step_callback
    )],
    output_path: Annotated[Path, typer.Option(
        "-o",
        "--output",
        file_okay=False,
        help="Folder with outputs"
    )] =Path(f"<step-name>-{datetime.now().isoformat(timespec='milliseconds')}"),
    input_folders: Annotated[list[Path], typer.Option(
        "-i",
        "--inputs",
        help="input folders",
        file_okay=False,
        exists=True,
    )] = [],
    executor: Annotated[str, typer.Option(
       # "",
        "-e",
        "--executor",
        help="executor to use",
        callback=executer_callback,
        autocompletion= lambda : ['BaseStepExecutor', 'PrometheusStepExecutor']
    )] = 'BaseStepExecutor',
    encapsulate_env: Annotated[bool, typer.Option(
    )] = True,
):
    """run"""
    output_path = Path(output_path.as_posix().replace("<step-name>", step.__name__))
    log.debug("executing run", extra={
        'parsed_args': {
            'step': step,
            'output_path': output_path,
            'input_folders': input_folders,
            'executor': executor,
            'encapsulate_env': encapsulate_env,
        }
    })
    return cmd_run(step, output_path, input_folders, executor, encapsulate_env)
@app.command(
    "inspect",
    no_args_is_help=True,
    help="Display information about a step"
)
def inspekt(
    step: Annotated[str, typer.Argument(
        allow_dash=False,
        help="module path to step",
        autocompletion=complete_step_import,
        callback=step_callback
    )],
    gen_env: Annotated[bool, typer.Option(
    )] = False
    ):
    """inspect"""
    return cmd_inspect(step, gen_env)
def update_log_level(log_level: str):
    """Fix for typer logs"""
    log_config = get_logging_dict_config(log_level)
    log_config['formatters']['default'] = {
        '()': "wurzel.cli.logger.WithExtraFormatter",
            'reduced': ["INFO"]
    }
    log_config['handlers']['default'] = {
        "()": "rich.logging.RichHandler",
        "formatter": "default",
        }
    logging.config.dictConfig(log_config)


@app.callback()
def main_args(
    verbose: Annotated[bool, typer.Option(
        "--verbose"
    )] = False,
    log_level: Annotated[str, typer.Option(
        "--log-level",
        autocompletion=lambda: ['CRITICAL','FATAL','ERROR','WARN','INFO']
    )] = "INFO"
):
    """global settings, main"""
    if not os.isatty(1):
        typer.core.rich = None
        logging.config.dictConfig(get_logging_dict_config(log_level))
        app.pretty_exceptions_enable = False
        app.pretty_exceptions_show_locals=False
    else:
        update_log_level(log_level)
        app.pretty_exceptions_enable = True
        app.pretty_exceptions_show_locals=True
        app.pretty_exceptions_short = not verbose
def main():
    """main"""
    app()



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import json

from pydantic import Field, validator
from qdrant_client.models import Distance
from wurzel.step.settings import StepSettings

class QdrantSettings(StepSettings):
    """All Settings related to Qdrant DB"""
    DISTANCE: Distance = Distance.DOT
    URI:str = "http://localhost:6333"
    COLLECTION: str
    COLLECTION_HISTORY_LEN: int = 10
    SEARCH_PARAMS: dict  = {"metric_type": "IP", "params": {}}
    INDEX_PARAMS: dict = {"index_type": "FLAT", "field_name": "vector"  , "distance": "Dot", "params": {}}
    APIKEY: Optional[str] = ""
    REPLICATION_FACTOR:int = Field(default=3, gt=0)
    BATCH_SIZE:int = Field(default=1024,gt=0)
    @validator('SEARCH_PARAMS', 'INDEX_PARAMS', pre=True)
    @classmethod
    def parse_json(cls, v):
        """validation for json"""
        if isinstance(v, str):
            return json.loads(v)
        return v



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from typing import (
    Generic, TypeVar,
    Type,
    Any, get_args
)
import pydantic
from pydantic_core import core_schema, PydanticCustomError
import pandera.typing as patyp
B = TypeVar('B',
            pydantic.BaseModel, None,
            patyp.DataFrame, list[pydantic.BaseModel],
            list[patyp.DataFrame], list[None])
class PathToFolderWithBaseModels(type(Path()), Generic[B]): #type: ignore[misc]
    """Used to Store both path information 
    as well as pydantic.BaseModel info.
    
    Inherits from Pathlib.Path:
    - Supports all path operations like / etc.
    
    
    #### Example:
    ```python
        class Person(pydantic.BaseModel):
            # Could be anything that inherits from BaseModel
            name: str
            age: int
        # Define new class to supply GenericType
        class JFP(PathToBaseModel[Person]): # JsonFilePath
            pass

        path = JFP("./max.json")
        # Load from path
        p_max = path.load_model()
        # Do anything
        p_max.age = p_max.age += 1
        # Store at path
        # Actual path can be modified like pathlib.Path
        path.save_model(p_max)
    ```
    """
    @classmethod
    def _type(cls) -> Type["PathToFolderWithBaseModels"]:
        """Get own type (used for pydantic)

        Raises:
            RuntimeError: if own type can't be found

        Returns:
            Type: PathToBaseModel
        """
        typ = getattr(cls, '__orig_bases__', [None])[0]
        if typ is None:
            raise RuntimeError(f"Could not get __orig_bases__ from class '{cls}'")
        return typ
    @classmethod
    def model_type(cls) -> Type[B]:
        """_summary_
        
        Raises:
            RuntimeError: if type can't be found
        Returns:
            Type[pydantic.BaseModel]: Type of Generic
        """
        try:
            return get_args(cls._type())[0]
        except Exception as err:
            raise RuntimeError("Model type could not be found") from err
    @classmethod
    def __get_pydantic_core_schema__(cls, _: Any, handler: pydantic.GetCoreSchemaHandler):
        return core_schema.with_info_after_validator_function(
            cls._validate_path,
            handler(cls._type()))
    @classmethod
    def _validate_path(cls, path: "PathToFolderWithBaseModels", _: core_schema.ValidationInfo) -> Path:
        if not path.is_dir():
            raise PydanticCustomError("path_not_dir", "Path is not a directory but a file")
        return cls(path)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .step import EmbeddingStep
from .huggingface import (
    HuggingFaceInferenceAPIEmbeddings,
    PrefixedAPIEmbeddings
)
__all__=["EmbeddingStep", "HuggingFaceInferenceAPIEmbeddings","PrefixedAPIEmbeddings"]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into Qdrant"""
#pylint: disable=duplicate-code
from logging import getLogger
import itertools
from pandera.typing import DataFrame
from qdrant_client import models
from wurzel.exceptions import StepFailed
from wurzel.step import TypedStep
from wurzel.steps.embedding.data import  EmbeddingMultiVectorResult
from wurzel.steps.qdrant.step import QdrantConnectorStep
from .data import QdranttMultiVectorResult
from .settings import QdrantSettings


log = getLogger(__name__)

def _batch(iterable, size):
    it = iter(iterable)
    while item := list(itertools.islice(it, size)):
        yield item
class QdrantConnectorMultiVectorStep(QdrantConnectorStep,
        TypedStep[QdrantSettings, DataFrame[EmbeddingMultiVectorResult], DataFrame[QdranttMultiVectorResult]]):
    """
    Qdrant connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings
    """
    def _create_collection(self, size: int):
        self.client.create_collection(
            collection_name=self.collection_name,
            vectors_config=models.VectorParams(
                size=size,
                distance=self.settings.DISTANCE,
                multivector_config=models.MultiVectorConfig(
                    comparator=models.MultiVectorComparator.MAX_SIM)
        ),
        replication_factor=self.settings.REPLICATION_FACTOR
        )
    def run(self, inpt: DataFrame[EmbeddingMultiVectorResult]) -> DataFrame[QdranttMultiVectorResult]:
        if not self.client.collection_exists(self.collection_name):
            self._create_collection(len(inpt["vectors"].loc[0][0]))
        df_result = self._insert_embeddings(inpt)
        return df_result



    def _insert_embeddings(self, data: DataFrame[EmbeddingMultiVectorResult])->DataFrame[QdranttMultiVectorResult]:

        log.debug(f"Creating Qdrant collection {self.collection_name}")


        log.info("Inserting embeddings", extra={
            'count': len(data),
            'collection': self.collection_name})
        points= [models.PointStruct(id=next(self.id_iter), #type: ignore[arg-type] # No MultiIndex, so always int.
                                    vector=row["vectors"],
                                    payload={"url":row["url"],
                                             "text":row["text"],
                                             "splits":row["splits"],
                                             "keywords":row["keywords"]}) for i, row in data.iterrows()]
        for point_chunk in _batch(points, self.settings.BATCH_SIZE):
            operation_info = self.client.upsert(collection_name=self.collection_name, wait= True,
                                            points=point_chunk)
            if operation_info.status != "completed":
                raise StepFailed(f"Failed to insert df chunk into collection '{self.collection_name}' {operation_info}")
            log.info("Successfully inserted vector_chunk",extra={
                'collection': self.collection_name,
                'count': len(point_chunk)
            })
        self._create_indices()
        self._update_alias()
        data = [{
                "text": entry.payload["text"],
                "url": entry.payload["url"],
                "splits": entry.payload["splits"],
                "vectors": entry.vector,
                "keywords":entry.payload["keywords"],
                "collection":self.collection_name,
                "id":entry.id} for entry in points]
        result = DataFrame[QdranttMultiVectorResult](data)
        return result



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from . import semantic_splitter
from .meta_settings import WZ, create_model
from .meta_steps import find_typed_steps_in_package
__all__=["semantic_splitter", "WZ", "create_model", "try_get_length", "find_typed_steps_in_package"]
def try_get_length(x) -> int:
    """Tries to get length, return 1 if fails

    Args:
        x (Any): get length of

    Returns:
        int: length or 1
    """
    try:
        return len(x)
    #pylint: disable=bare-except
    except:
        return 1



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into milvus"""

from logging import getLogger
import pandas as pd
from pandera.typing import DataFrame
from pymilvus import (
    FieldSchema,
    CollectionSchema,
    DataType
)

from pymilvus.exceptions import MilvusException
from pymilvus.milvus_client.index import IndexParams
from pymilvus.milvus_client import MilvusClient

from wurzel.step import TypedStep
from wurzel.exceptions import NoPreviousCollection, StepFailed
from wurzel.steps.embedding.data import EmbeddingResult

from .data import Result as MilvusResult
from .settings import MilvusSettings
log = getLogger(__name__)

class MilvusConnectorStep(TypedStep[MilvusSettings,DataFrame[EmbeddingResult], MilvusResult]): # pragma: no cover
    """
    Milvus connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings
    """
    milvus_timeout:float = 20.0
    def __init__(self) -> None:
        super().__init__()
        # milvus stuff passed as environment
        # because we need to enject them into the DVC step during runtime,
        # not during DVC pipeline definition time
        uri = f"http://{self.settings.HOST}:{self.settings.PORT}"
        if not self.settings.PASSWORD or not self.settings.USER:
            log.warning("MILVUS_HOST, MILVUS_USER or MILVUS_PASSWORD for Milvus not provided."
                            " Thus running in non-credential Mode")
        self.client:MilvusClient = MilvusClient(uri=uri,
                                                user=self.settings.USER,
                                                password=self.settings.PASSWORD,
                                                timeout=self.milvus_timeout)
        self.collection_index:IndexParams = IndexParams(**self.settings.INDEX_PARAMS)
        self.collection_history_len = self.settings.COLLECTION_HISTORY_LEN

        self.collection_prefix = self.settings.COLLECTION

    def __del__(self):
        if getattr(self,'client', None):
            self.client.close()
    def run(self, inpt: DataFrame[EmbeddingResult]) -> MilvusResult:
        self._insert_embeddings(inpt)
        try:
            old = self.__construct_last_collection_name()
        except NoPreviousCollection:
            old = ""
        self._retire_collection()
        return MilvusResult(new=self.__construct_current_collection_name(),old=old)

    def _insert_embeddings(self, data: pd.DataFrame):

        collection_name = self.__construct_next_collection_name()
        log.info(f"Creating milvus collection {collection_name}")
        collection_schema = CollectionSchema(
            fields=[
                FieldSchema(name="pk", dtype=DataType.INT64,
                            is_primary=True, auto_id=True),
                FieldSchema(name="text", dtype=DataType.VARCHAR,
                            max_length=3000),
                FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=len(
                    data["vector"].loc[0])),
                FieldSchema(name="url", dtype=DataType.VARCHAR, max_length=300)
            ],
            description="Collection for storing Milvus embeddings"
        )

        log.info("schema created")
        self.client.create_collection(collection_name=collection_name, schema=collection_schema)
        log.info("collection created")
        log.info(f"Inserting embedding {len(data)} into collection {collection_name}")
        result:dict = self.client.insert(collection_name=collection_name, data=data.to_dict("records"))
        if result["insert_count"] != len(data):
            raise StepFailed(f"Failed to insert df into collection '{collection_name}'."
                            f'{result["insert_count"]}/{len(data)} where successful')
        log.info(f"Successfully inserted {len(data)} vectors"
                    f" into collection '{collection_name}'")
        self.client.create_index(collection_name=collection_name,
                                index_params=self.collection_index)
        log.info(f"Successfully craeted index {self.collection_index}"
                    f" into collection '{collection_name}")
        self.client.load_collection(collection_name)
        log.info(f"Successfully loaded the collection {collection_name}'"
                f" into collection '{collection_name}'")
        try:
            self.client.release_collection(self.__construct_last_collection_name())
        except NoPreviousCollection:
            pass
        self._update_alias(collection_name)



    def _retire_collection(self):
        collections_versioned:dict[int,str] = self._get_collection_versions()
        to_delete  = sorted(collections_versioned.keys())[:-self.collection_history_len]
        if not to_delete :
            return

        for col_v in to_delete:
            col = collections_versioned[col_v]
            log.info(f"deleting {col} collection caused by retirement")
            self.client.drop_collection(col, timeout=self.milvus_timeout)



    def _update_alias(self, collection_name):
        try:
            self.client.create_alias(
                collection_name=collection_name,
                alias=self.collection_prefix, timeout=self.milvus_timeout
            )
        except MilvusException:
            self.client.alter_alias(
                collection_name=collection_name,
                alias=self.collection_prefix, timeout=self.milvus_timeout
            )

    def __construct_next_collection_name(self) -> str:
        previous_collections =self._get_collection_versions()
        if not previous_collections:
            return f"{self.collection_prefix}_v1"
        previous_version = max(previous_collections.keys())
        log.info(f"Found version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version+1}"

    def __construct_last_collection_name(self) -> str:
        previous_collections =self._get_collection_versions()
        if not previous_collections or len(previous_collections) <=1:
            raise NoPreviousCollection(f"Milvus does not contain a previous collection for {self.collection_prefix}")
        previous_version = sorted(previous_collections.keys())[-2]
        log.info(f"Found previous version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version}"

    def __construct_current_collection_name(self) -> str:
        previous_collections =self._get_collection_versions()
        if not previous_collections or len(previous_collections) <1:
            raise NoPreviousCollection(f"Milvus does not contain a previous collection for {self.collection_prefix}")
        previous_version = sorted(previous_collections.keys())[-1]
        log.info(f"Found previous version v{previous_version}")
        return f"{self.collection_prefix}_v{previous_version}"

    def _get_collection_versions(self)->dict[int,str]:
        previous_collections = self.client.list_collections(timeout=self.milvus_timeout)
        versioned_collections = {int(previous.split("_v")[-1]): previous for previous in previous_collections
                            if self.collection_prefix in previous}
        return versioned_collections



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .datacontract import PanderaDataFrameModel, PydanticModel, DataModel
from .common import MarkdownDataContract
__all__=["PanderaDataFrameModel", "PydanticModel", "MarkdownDataContract", "DataModel"]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pandera.typing import Series
from pandera import Field
from wurzel.datacontract import PanderaDataFrameModel

class EmbeddingResult(PanderaDataFrameModel):
    """data contract of embedding output"""
    text: Series[str]
    url: Series[str] = Field(nullable=True, default=" ", coerce=True, description="url used for search ")
    vector: Series[list[float]]
    keywords: Series[str]= Field(nullable=True, default=" ", coerce=True, description="Keywords used for search ")


class EmbeddingMultiVectorResult(PanderaDataFrameModel):
    """data contract of embedding output"""
    text: Series[str]
    url: Series[str] = Field(nullable=True, default=" ", coerce=True, description="url used for search ")
    vectors: Series[list[list[float]]]
    keywords: Series[str]= Field(nullable=True, default=" ", coerce=True, description="Keywords used for search ")
    splits: Series[list[str]]= Field(nullable=True, default=None, coerce=True, description="splits of the Multivector")



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Self
from pathlib import Path
from re import (
    compile as _re_compile,
    Pattern as _re_pattern)
import pydantic

from .datacontract import PydanticModel

_RE_TOPIC = _re_compile(r'topics:\s*(.*)')
_RE_URL = _re_compile(r'url:\s*(.*)')
_RE_BODY = _re_compile(r'---[\s\S]*?---\s*([\s\S]*)')

class MarkdownDataContract(PydanticModel):
    """contract of the input of the EmbeddingStep"""
    md : str
    keywords: str
    url: str # Url of pydantic is buggy in serialization

    @classmethod
    @pydantic.validate_call
    def from_dict_w_function(cls, doc: dict[str, Any], func: Callable[[str],str]):
        """Create a MarkdownDataContract from a dict and apply a custom func to test"""
        return cls(
            md=func(doc["text"]),
            url=doc["metadata"]["url"],
            keywords=doc["metadata"]["keywords"]
        )
    @classmethod
    def from_file(cls, path: Path, url_prefix: str = "") -> Self:
        """Load MdContract from .md file

        Args:
            path (Path): Path to file

        Returns:
            MarkdownDataContract: The file that was loaded
        """
        def find_first(pattern: _re_pattern, text: str, fallback: str):
            x = pattern.findall(text)
            return x[0] if len(x) >= 1 else fallback
        md = path.read_text()
        return MarkdownDataContract(
            md=str(find_first(_RE_BODY, md, md)),
            url=str(find_first(_RE_URL,md, url_prefix + path.as_posix())),
            keywords=str(find_first(_RE_TOPIC, md, path.name.split(".")[0])))



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from wurzel.datacontract import PydanticModel

class Result(PydanticModel):
    """dummy output because its required to have a output data contract """
    old: str
    new: str



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .base_executor import BaseStepExecutor
from .prometheus_executor import PrometheusStepExecutor



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import abc
from ast import literal_eval
from pathlib import Path
from typing import (
    Union, Self,
    get_origin, Type
)
import json
import pydantic
import pandas as pd
import pandera as pa
import pandera.typing as patyp

class DataModel:
    """interface definition of a Contract model
    Contains method to store and load to a path
    """
    @classmethod
    @abc.abstractmethod
    def save_to_path(cls, path: Path, obj: Union[Self, list[Self]]) -> Path:
        """abstract function to save the obj at the given path"""
    @classmethod
    @abc.abstractmethod
    def load_from_path(cls, path: Path, *args) -> Self:
        """abstract function to load the data from the given Path"""

class PanderaDataFrameModel(pa.DataFrameModel,DataModel):
    """Data Model contract specified with pandera
    Using Panda Dataframe. Mainly for CSV shaped data"""
    @classmethod
    def save_to_path(cls, path: Path, obj: Union[Self, list[Self]]) -> Path:
        path = path.with_suffix(".csv")
        if not isinstance(obj, pd.DataFrame):
            raise NotImplementedError(f"Cannot store {type(obj)}")
        obj.to_csv(path,index=False)
        return path
    @classmethod
    def load_from_path(cls, path: Path, *args) -> Self:
        """switch case to find the matching file ending"""
        read_data = pd.read_csv(path.open(encoding="utf-8"))
        for key,atr in cls.to_schema().columns.items():
            if atr.dtype.type == list:
                read_data[key] = read_data[key].apply(literal_eval)
        return patyp.DataFrame[cls](read_data)
class PydanticModel(pydantic.BaseModel,DataModel):
    """DataModel contract specified with pydantic"""
    @classmethod
    def save_to_path(cls, path: Path, obj: Union[Self, list[Self]]):
        """Wurzel save model

        Args:
            path (Path): location
            obj (Union[Self, list[Self]]): obj(s) to store

        Raises:
            NotImplementedError
        """
        path = path.with_suffix(".json")
        if isinstance(obj, list):
            with path.open("wt", encoding="UTF-8") as fp:
                json.dump(obj,fp, default=pydantic.BaseModel.model_dump)
        elif isinstance(obj, cls):
            with path.open("wt", encoding="UTF-8") as fp:
                fp.write(obj.model_dump_json())
        else:
            raise NotImplementedError(f"Cannot store {type(obj)}")
        return path
    #pylint: disable=arguments-differ
    @classmethod
    def load_from_path(cls, path: Path, model_type: Type[Union[Self, list[Self]]]) -> Union[Self, list[Self]]:
        """Wurzel load model

        Args:
            path (Path): load model from
            model_type (Type[Union[Self, list[Self]]]): expected type

        Raises:
            NotImplementedError

        Returns:
            Union[Self, list[Self]]: dependent on expected type
        """
        if get_origin(model_type) is None:
            if issubclass(model_type, pydantic.BaseModel):
                return cls(**json.load(path.open(encoding="utf-8")))
        elif get_origin(model_type) == list:
            data = json.load(path.open(encoding="utf-8"))
            for i, entry in enumerate(data):
                data[i] = cls(**entry)
            return data

        raise NotImplementedError(f"Can not load {model_type}")



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Type
from pathlib import Path
from wurzel import TypedStep

def main(step: Type[TypedStep],
         output_path: Path,
         input_folders: set[Path],
         executor_str_value: str,
         encapsulate_env=True,):
    """Execute"""
    executor = executor_str_value
    with executor(dont_encapsulate=not encapsulate_env) as ex:
        ex(step, input_folders, output_path)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .step.settings import (
    StepSettings as Settings,
    NoSettings
)
from .step import TypedStep
from .datacontract import PanderaDataFrameModel, PydanticModel, MarkdownDataContract
from .path import PathToFolderWithBaseModels
from .step_executor import BaseStepExecutor, PrometheusStepExecutor
#pylint: disable-next=invalid-name

__all__=[
    "TypedStep", "PanderaDataFrameModel", "PydanticModel", "MarkdownDataContract",
    "PathToFolderWithBaseModels",
    "steps", "step_executor", "NoSettings", "Settings",
    "BaseStepExecutor", "PrometheusStepExecutor"
]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""
consists of DVCSteps to embedd files and save them as for example as csv
"""
# Standard library imports
from logging import getLogger
from typing import TypedDict

import numpy as np
from pandera.typing import DataFrame
from joblib import Parallel, delayed


from wurzel.step import TypedStep
from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import SplittException, StepFailed, EmbeddingAPIException
from wurzel.steps.embedding import EmbeddingStep
# Local application/library specific imports
from .settings import EmbeddingSettings


from .data import EmbeddingMultiVectorResult

log = getLogger(__name__)



class _EmbeddedMultiVector(TypedDict):
    "dict definition of a embedded document"
    text:str
    url:str
    vectors: np.ndarray
    splits: list[str]



class EmbeddingMultiVectorStep(EmbeddingStep,
            TypedStep[EmbeddingSettings,list[MarkdownDataContract], DataFrame[EmbeddingMultiVectorResult]]):
    """
    Step for consuming list[MarkdownDataContract]
    and returning DataFrame[EmbeddingMultiVectorResult]
    """
    def run(self, inpt: list[MarkdownDataContract]) -> DataFrame[EmbeddingMultiVectorResult]:
        """ Executes the embedding step by processing input markdown files, generating embeddings,
        and saving them to a CSV file.
        """
        def process_document(doc):
            try:
                return self._get_embedding(doc)
            except EmbeddingAPIException as err:
                log.warning(f"Skipped because EmbeddingAPIException: {err.message}", extra={"markdown": str(doc)})
                return None

        results = Parallel(
            backend="threading",n_jobs=self.settings.N_JOBS)(delayed(process_document)(doc) for doc in inpt)

        rows = [res for res in results if res is not None]
        failed = len(results) - len(rows)

        if failed:
            log.warning(f"{failed}/{len(inpt)} got skipped")
        if failed == len(inpt):
            raise StepFailed(f"All {len(inpt)} embeddings got skipped")

        return DataFrame[EmbeddingMultiVectorResult](rows)

    def _get_embedding(self, doc: MarkdownDataContract)->_EmbeddedMultiVector:
        """
        Generates an embedding for a given text and context.

        Parameters
        ----------
        d : dict
            A dictionary containing the text and context for which to generate the embedding.

        Returns
        -------
        dict
            A dictionary containing the original text, its embedding, and the source URL.
        """
        def prepare_plain(document: MarkdownDataContract)->str:
            plain_text = self.markdown.convert(document.md)
            plain_text = self._replace_link(plain_text)
            return plain_text
        try:
            splitted_md_rows = self._split_markdown([doc])
        except SplittException as err:
            raise EmbeddingAPIException("splitting failed") from err
        vectors = [self.embedding.embed_query(prepare_plain(split)) for split in splitted_md_rows ]
        if not vectors:
            raise EmbeddingAPIException("Embedding failed for all splits")

        context = self.get_simple_context(doc.keywords)

        return {"text": doc.md,
                "vectors":vectors,
                "url":doc.url,
                "keywords": context, 
                "splits":[split.md for split in splitted_md_rows]
        }



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""Base Executor"""
from contextvars import copy_context
from pathlib import Path
from typing import Generator, Optional, Type, Self, Union, Callable, Any
from types import NoneType
from contextlib import contextmanager
import time
import json
from logging import getLogger
import os
import pydantic
from pydantic_core import ValidationError
import pandera.typing as patyp
from asgi_correlation_id import correlation_id

from wurzel.exceptions import (
    ContractFailedException,
    EnvSettingsError,
    LoggedCustomException,
    StepFailed)
from wurzel.step.typed_step import (
    TypedStep,
)
from wurzel.step.history import (
    History,
    step_history,
)
from wurzel.datacontract import PanderaDataFrameModel, PydanticModel, datacontract
from wurzel.path import PathToFolderWithBaseModels
from wurzel.utils import WZ, create_model, try_get_length
log = getLogger(__name__)
class StepReport(pydantic.BaseModel):
    """Generated by TypedStep Class"""
    model_config = pydantic.ConfigDict(
        frozen=True
    )
    inputs: int = 0
    results: int = 1
    time_to_load: float
    time_to_execute: float
    time_to_save: float
    step_name: str
    history: list[str]

@contextmanager
def step_env_encapsulation(step_cls: Type[TypedStep]):
    """Create a virtual env and remove it afterwards"""
    old = os.environ.copy()
    name = step_cls.__name__.upper()
    step = WZ(step_cls)
    if BaseStepExecutor.is_allow_extra_settings():
        log.info("Allowing extra settings fields")
    if step.settings_class != NoneType:
        settings = create_model(
            [step],
            allow_extra_fields=BaseStepExecutor.is_allow_extra_settings()
        )
        try:
            inner_settings: pydantic.BaseModel = getattr(settings(), name)
        except ValidationError as err:
            e =  EnvSettingsError(
                f"could not create {step.settings_class.__name__} from env for {step_cls.__name__}"
            )
            e.add_note("To fix these issues setup env vars in the format <step_name>__<var>")
            raise e from err
        settings_dict = pydantic.BaseModel.model_dump(
            inner_settings, mode="json",
            exclude_none=True, exclude_unset=True
        )
        for field_name, value in settings_dict.items():
            if isinstance(value, (list, dict, set, tuple)):
                value = json.dumps(value)
            else:
                value = str(value)
            os.environ[field_name] = value
        log.info("Running with encapsulated settings", extra={
            'settings': inner_settings.model_dump(mode="json")
        })
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old)

class BaseStepExecutor():
    """Execution Environment for Typed KP Steps
    
    An Executer takes care of loading & storing data from `PathToBaseModel`.
    The Executor makes use of strong & enforced types in combination with the
    KP step itself.

    The Executor is responsible for running single KP Steps. However in the 
    future this might get extendend.


    ### Usage:

    The Executor can be used as a context manager.
    ```python
    with BaseStepExecutor() as exc:
        exc(MyStep, [], Path("output"))
        # exc(...) == exc.execute_step(...)
    ``` 
    
    
    """
    __dont_encapsulate: bool
    @classmethod
    def is_allow_extra_settings(cls: "BaseStepExecutor") -> bool:
        """Checks os for ALLOW_EXTRA_SETTINGS

        Args:
            cls (BaseStepExecutor)

        Returns:
            bool: ALLOW_EXTRA_SETTINGS
        """
        return os.environ.get("ALLOW_EXTRA_SETTINGS", "False").upper() == "TRUE"
    def __init__(self, dont_encapsulate: bool = False) -> None:
        self.__dont_encapsulate = dont_encapsulate
    def store(
            self,
            step: TypedStep,
            hist: History,
            obj: Union[PanderaDataFrameModel, PydanticModel], path: PathToFolderWithBaseModels)->Path:
        """Store step result

        Args:
            step (TypedStep): Step Object
            obj (Union[PanderaDataFrameModel, PydanticModel]): Result of step
            path (PathToBaseModel): Where to save
        """
        if not path.is_dir():
            path.mkdir()
        output_model_class: Type[datacontract.DataModel] = step.output_model_class
        return output_model_class.save_to_path(path / f"{hist}", obj)

    def load(
            self,
            step: TypedStep,
            path: PathToFolderWithBaseModels
            ):
        """Load step input

        Args:
            step (TypedStep): step obj for which to load input
            path (PathToBaseModel): Path to input
            model_type (Union[Type, None]): Type

        Returns:
            Union[Type[None], Union[PanderaDataFrameModel, PydanticModel]]: _description_
        """
        input_model_class: Type[datacontract.DataModel] = step.input_model_class
        for p in path.glob("*"):
            start = time.time()
            data = input_model_class.load_from_path(p, step.input_model_type)
            yield (data, History('.'.join(p.name.split(".")[:-1]), step)), time.time()-start
    def __load_data(
            self,
            step: TypedStep,
            inputs: set[Union[PydanticModel, patyp.DataFrame[PanderaDataFrameModel], PathToFolderWithBaseModels]]
            )  -> Generator[tuple[
                tuple[
                    Union[PydanticModel, patyp.DataFrame[PanderaDataFrameModel], list[PydanticModel], None],
                    History],
                float], Any, NoneType]:
        """Load (if necessary from disk) inputdata

        Args:
            step (TypedStep): step to load data for
            inputs (set[Union[PydanticModel, DataFrame[PanderaDataFrameModel], PathToFolderWithBaseModels]]): inputs

        Yields:
            Loaded input: With history already extended
        """
        no_inputs_supplied = inputs in [[], set()]
        if no_inputs_supplied:
            # Only yield once
            no_inputs_supplied = False
            yield (None, History(step)), 0
        for inpt in inputs:
            if isinstance(inpt, (datacontract.DataModel, PydanticModel, patyp.DataFrame)):
                yield (inpt, History("[Memory]",step)), 0
            elif isinstance(inpt, (Path, PathToFolderWithBaseModels)):
                for (inpt, hist), took in self.load(step, inpt):
                    yield (inpt, hist), took
            else:
                raise NotImplementedError(
                    f"Cannot load/convert {inpt} as input for a step"
                )
    def __execute_step(
            self,
            step_cls: Type[TypedStep],
            inputs: set[PathToFolderWithBaseModels],
            output_path: Optional[PathToFolderWithBaseModels]):
        """Exceute specified step"""
        step = step_cls()
        # pylint: disable=protected-access
        if output_path:
            output_path = step._internal_output_class(output_path)
        run: Callable[[list], Any] = pydantic.validate_call(
            step.run, validate_return=True)
        was_called_once = False
        try:
            for (inpt, history), load_time in self.__load_data(step, inputs):
                was_called_once = True
                log.info(f"Start: {step_cls.__name__}.run({history[:-1]}) -> {output_path}",)
                history: History
                run_start = time.time()
                ctx = copy_context()
                token = step_history.set(history)
                ctx.run(step_history.set, history)
                try:
                    res = ctx.run(run, inpt)
                finally:
                    step_history.reset(token)
                run_time = time.time()-run_start
                store_time = 0
                if output_path:
                    self.store(step, history, res, output_path)
                    store_time = time.time()-run_start
                report = StepReport(
                    time_to_load=load_time,
                    time_to_execute=run_time,
                    time_to_save=store_time,
                    step_name=step_cls.__name__,
                    inputs=try_get_length(inpt),
                    results = getattr(res, '__len__', lambda : 1)(),
                    history=history.get()
                )
                log.info(f"Finished: {step_cls.__name__}.run({history[:-1]}) -> {output_path}",
                    extra=report.model_dump())
                yield res, report
            log.info(f"{step_cls.__name__}.finalize()")
            if not was_called_once:
                log.error("Run was never executed", extra={
                    'step': step.__class__.__name__,
                    'inpts': inputs,
                })
            step.finalize()
        except ValidationError as err:
            raise ContractFailedException(
                f"{inputs} does not conform the data contract of {step.input_model_class.__name__}") from err


    def execute_step(
            self,
            step_cls: Type[TypedStep],
            inputs: Optional[set[PathToFolderWithBaseModels]],
            output_dir: Optional[PathToFolderWithBaseModels]) -> list[tuple[Any, StepReport]]:
        """Execute a step

        Args:
            step_cls (Type[TypedStep]): class to execute
            inputs (Optional[set[PathToFolderWithBaseModels]]): Either objects or a folder path
            output_dir (Optional[PathToFolderWithBaseModels]): If folder is supllied, save to folder
        
        Returns:
            list[tuple[Any, StepReport]]: Result of step and repots in list
        """
        if not inputs:
            inputs = set()
        try:
            correlation_id.set(step_cls.__name__)
            log.info(f"{self.__class__.__name__} - start: {step_cls.__name__}")
            if self.__dont_encapsulate:
                return list(self.__execute_step(step_cls, inputs, output_dir))
            with step_env_encapsulation(step_cls):
                return list(self.__execute_step(step_cls, inputs, output_dir))
        except LoggedCustomException:
            raise
        except Exception as e:
            raise StepFailed(f"{self.__class__.__name__} error during {step_cls.__name__}: {str(e)}") from e
        finally:
            log.info(f"{self.__class__.__name__} - done: {step_cls.__name__}")
            correlation_id.set(None)
    __call__ = execute_step

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_details):
        pass



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""contains custom logged exceptions"""

import logging
log = logging.getLogger(__name__)


class LoggedCustomException(Exception):
    """custom logged exception class"""

    def __init__(self, message):
        """constructor"""
        self.message = message
        super().__init__(message)
        log.exception(self)

    def __repr__(self) -> str:
        return f"{str(self.__class__.__name__)}: {self.message}"

    def __str__(self) -> str:
        return self.__repr__()
class EnvSettingsError(ValueError):
    """Special Error for env related errors in regards to TypedStep settings"""

class BadResponse(LoggedCustomException):
    """raises if an HTTP api does not return 200
    """


class StepFailed(LoggedCustomException):
    """raised when a step fails unintented"""


class InvalidCountryCodeException(LoggedCustomException):
    """raised when countryCode in url path is not recognized
    or not in accordance to ISO_3166"""

class ContractFailedException(LoggedCustomException):
    """raised when a data contract is not fulfilled"""
class StaticTypeError(ContractFailedException, TypeError):
    """raised if static type hinting are wrong"""

class ConvertFailed(LoggedCustomException):
    """raised when html2md fails"""

class CLIException(LoggedCustomException):
    """raised when the given string is not a valid Callable Pipeline"""

class NoPreviousCollection(LoggedCustomException):
    """raised when milvus does not contain previous collection"""

class MarkdownException(LoggedCustomException):
    """raised when Markdown libary complains"""
class CustomQdrantException(LoggedCustomException):
    """raised when qdrant status is not fine"""

class SplittException(LoggedCustomException):
    """ Document had to have children """

class EmbeddingException(LoggedCustomException):
    """The EmbeddingService returned data causing errors"""


class EmbeddingAPIException(LoggedCustomException):
    """The EmbeddingServiceAPI experienced an Error"""

class UnrecoverableFatalException(LoggedCustomException):
    """Raised when the application encountered a problem that cannot be recovered"""



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .settings import MilvusSettings
__all__ = ["MilvusSettings"]
try:
    _HAS_MILVUS = True
    import pymilvus as _
except ImportError:
    _HAS_MILVUS = False
if _HAS_MILVUS:
    from .step import MilvusConnectorStep
    __all__ = [*__all__, "MilvusConnectorStep"]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import inspect
from pathlib import Path
from typing import Any, Type, TypedDict

import yaml
from wurzel import BaseStepExecutor, PrometheusStepExecutor, TypedStep

import wurzel
import wurzel.cli

class DvcDict(TypedDict):
    """Internal Representation"""
    cmd: str
    deps: list[str]
    outs: list[str]
    always_changed: bool

class DvcAdapter():
    """'Adapter' which creates DVC yamls"""
    @classmethod
    def generate_dict(cls,
        step: TypedStep,
        data_output_folder: Path,
        executor: Type[BaseStepExecutor] = PrometheusStepExecutor,
        encapsulate_env: bool = True) -> dict[str, DvcDict]:
        """generates the resulting dvc.yaml as dict by calling all
        its required steps as well, in recursive manner.

        Returns
        -------
        dict
            _description_
        """
        if not isinstance(data_output_folder, Path):
            data_output_folder = Path(data_output_folder)
        result: dict[str, Any] = {}
        outputs_of_deps: list[Path] = []
        for o_step in step.required_steps:
            dep_result = cls.generate_dict(o_step,data_output_folder)
            result |= dep_result
            outputs_of_deps += dep_result[o_step.__class__.__name__]["outs"]
        output_path =  data_output_folder / step.__class__.__name__
        cmd = wurzel.cli.generate_cli_call(
            step.__class__,
            inputs=outputs_of_deps,
            output=output_path,
            executor=executor,
            encapsulate_env=encapsulate_env,
        )
        return result | {step.__class__.__name__: {
                "cmd": cmd,
                "deps": [inspect.getfile(step.__class__), *outputs_of_deps],
                "outs": [output_path],
                "always_changed": step.is_leaf()
            }}
    @classmethod
    def generate_yaml(cls, step, path: Path, data_output_folder: Path):
        """generates the dvc.yaml"""
        with open(path, "w", encoding="utf-8") as file:
            data = cls.generate_dict(step, data_output_folder)
            for k in data:
                for key in ["outs", "deps"]:
                    data[k][key] = [str(p) for p in  data[k][key]]
            yaml.dump({"stages": data}, file)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import logging.config

from wurzel.utils.logging import JsonFormatter



class WithExtraFormatter(JsonFormatter):
    """Custom formatter with some structured logging support"""
    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        json_part = self._get_output_dict(record)
        msg = json_part.pop("message")
        if self.reduced_levels and record.levelno in self.reduced_levels:
            json_part.pop('thread', None)
            json_part.pop('threadName', None)
            json_part.pop('processName', None)
            json_part.pop('process', None)
        json_part.pop('level', None)
        json_part.pop('@timestamp', None)
        json_part.pop('file', None)
        exc_text = json_part.pop("exc_text", "")
        return " ".join([
            f"'{msg}'" + (f" : {json.dumps(json_part)}" if json_part else "") + exc_text
        ])



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .path import PathToFolderWithBaseModels

__all__=["PathToFolderWithBaseModels"]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

#pragma: no cover
#pylint: skip-file
"""CLI Entry"""
from wurzel.cli._main import main as _main

if __name__ == '__main__':
    _main()



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into Qdrant"""
#pylint: disable=duplicate-code
from logging import getLogger
import itertools
from pandera.typing import DataFrame
from qdrant_client import QdrantClient, models
from wurzel.exceptions import StepFailed, CustomQdrantException
from wurzel.step import TypedStep, step_history
from wurzel.steps.embedding.data import EmbeddingResult
from .data import QdrantResult
from .settings import QdrantSettings


log = getLogger(__name__)

def _batch(iterable, size):
    it = iter(iterable)
    while item := list(itertools.islice(it, size)):
        yield item
class QdrantConnectorStep(TypedStep[QdrantSettings, DataFrame[EmbeddingResult], DataFrame[QdrantResult]]):
    """
    Qdrant connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings
    """
    _timeout:int = 20
    s: QdrantSettings
    client: QdrantClient
    collection_name: str
    def __init__(self) -> None:
        super().__init__()
        # Qdrant stuff passed as environment
        # because we need to enject them into the DVC step during runtime,
        # not during DVC pipeline definition time
        # uri = ":memory:"
        log.info(f"connecting to {self.settings.URI}")
        if not self.settings.APIKEY:
            log.warning("QDRANT__APIKEY for Qdrant not provided."
                            " Thus running in non-credential Mode")
        self.client = QdrantClient(location=self.settings.URI, api_key=self.settings.APIKEY, timeout=self._timeout)
        self.collection_name = self.__construct_next_collection_name()
        self.id_iter = self.__id_gen()
    def __del__(self):
        if getattr(self,'client', None):
            self.client.close()
    def finalize(self) -> None:
        self._create_indices()
        self._update_alias()
        self._retire_collections()
        return super().finalize()
    def __id_gen(self):
        i = 0
        while True:
            i += 1
            yield i
    def run(self, inpt: DataFrame[EmbeddingResult]) -> DataFrame[QdrantResult]:
        if not self.client.collection_exists(self.collection_name):
            self._create_collection(len(inpt["vector"].loc[0]))
        df_result = self._insert_embeddings(inpt)
        return df_result

    def _create_collection(self, size: int):
        log.debug(f"Creating Qdrant collection {self.collection_name}")
        self.client.create_collection(
            collection_name=self.collection_name,
            vectors_config=models.VectorParams(
                size=size,
                distance=self.settings.DISTANCE),
            replication_factor=self.settings.REPLICATION_FACTOR
        )
    def _insert_embeddings(self, data: DataFrame[EmbeddingResult]):
        log.info("Inserting embeddings", extra={
            'count': len(data),
            'collection': self.collection_name})
        points= [models.PointStruct(id=next(self.id_iter), #type: ignore[arg-type] # No MultiIndex, so always int.
                                    vector=row["vector"],
                                    payload={
                                        "url":row["url"],
                                        "text":row["text"],
                                        "keywords":row["keywords"],
                                        "history": str(step_history.get())}) for _, row in data.iterrows()]
        for point_chunk in _batch(points, self.settings.BATCH_SIZE):
            operation_info = self.client.upsert(collection_name=self.collection_name, wait= True,
                                            points=point_chunk)
            if operation_info.status != "completed":
                raise StepFailed(f"Failed to insert df chunk into collection '{self.collection_name}' {operation_info}")
            log.info("Successfully inserted vector_chunk",extra={
                'collection': self.collection_name,
                'count': len(point_chunk)
            })
        data = [{
                "text": entry.payload["text"],
                "url": entry.payload["url"],
                "vector": entry.vector,
                "history": entry.payload["history"],
                "keywords":entry.payload["keywords"],
                "collection":self.collection_name,
                "id":entry.id} for entry in points]
        result = DataFrame[QdrantResult](data)
        return result

    def _create_indices(self):

        self.client.create_payload_index(
                    collection_name=self.collection_name,
                    field_name="keywords",
                    field_schema=models.TextIndexParams(type=models.TextIndexType.TEXT,
                                                        tokenizer=models.TokenizerType.WHITESPACE))
        self.client.create_payload_index(
                    collection_name=self.collection_name,
                    field_name="url",
                    field_schema=models.TextIndexParams(type=models.TextIndexType.TEXT,
                                                        tokenizer=models.TokenizerType.PREFIX,
                                                        min_token_len=3))
        self.client.create_payload_index(
                    collection_name=self.collection_name,
                    field_name="text",
                    field_schema=models.TextIndexParams(type=models.TextIndexType.TEXT,
                                                        tokenizer=models.TokenizerType.MULTILINGUAL))
        self.client.create_payload_index(
                    collection_name=self.collection_name,
                    field_name="history",
                    field_schema=models.TextIndexParams(type=models.TextIndexType.TEXT,
                                                        tokenizer=models.TokenizerType.WORD))

    def _retire_collections(self):
        collections_versioned:dict[int,str] = self._get_collection_versions()
        to_delete  = list(collections_versioned.keys())[:-self.settings.COLLECTION_HISTORY_LEN]
        if not to_delete :
            return

        for col_v in to_delete:
            col = collections_versioned[col_v]
            log.info(f"deleting {col} collection caused by retirement")
            self.client.delete_collection(col)



    def _update_alias(self):
        success = self.client.update_collection_aliases(
            change_aliases_operations=[
                models.CreateAliasOperation(
                    create_alias=models.CreateAlias(
                        collection_name=self.collection_name,
                        alias_name=self.settings.COLLECTION
        ))])
        if not success:
            raise CustomQdrantException("Alias Update failed")

    def __construct_next_collection_name(self) -> str:
        previous_collections =self._get_collection_versions()
        if not previous_collections:
            return f"{self.settings.COLLECTION}_v1"
        previous_version = max(previous_collections.keys())
        log.info(f"Found version v{previous_version}")
        return f"{self.settings.COLLECTION}_v{previous_version+1}"

    def _get_collection_versions(self)->dict[int,str]:

        previous_collections = self.client.get_collections().collections
        versioned_collections = {int(previous.name.split("_v")[-1]): previous.name for previous in previous_collections
                            if self.settings.COLLECTION in previous.name}
        return dict(sorted(versioned_collections.items()))



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""
consists of DVCSteps to embedd files and save them as for example as csv
"""
# Standard library imports
from logging import getLogger



from pydantic import Field

from wurzel import Settings

from wurzel.step import TypedStep
from wurzel.datacontract import MarkdownDataContract
from wurzel.exceptions import MarkdownException, SplittException
from wurzel.utils.semantic_splitter import SemanticSplitter
# Local application/library specific imports

class SplitterSettings(Settings):
    """Anything Embedding-related"""

    BATCH_SIZE:int = Field(100, gt=0)
    TOKEN_COUNT_MIN:int = Field(64, gt=0)
    TOKEN_COUNT_MAX:int = Field(1024, gt=1)
    TOKEN_COUNT_BUFFER:int = Field(32, gt=0)

log = getLogger(__name__)



class SimpleSplitterStep(TypedStep[SplitterSettings, list[MarkdownDataContract], list[MarkdownDataContract]]):
    """SimpleSplitterStep to split Markdown Documents rundimentory in medium size chunks"""
    def __init__(self) -> None:
        super().__init__()
        self.splitter = SemanticSplitter(token_limit=self.settings.TOKEN_COUNT_MAX,
                                             token_limit_buffer=self.settings.TOKEN_COUNT_BUFFER,
                                             token_limit_min=self.settings.TOKEN_COUNT_MIN)

    def run(self, inpt: list[MarkdownDataContract]) -> list[MarkdownDataContract]:
        """ Executes the split step by processing input markdown files, generating smaller splitted markdown files,
        by preserving the headline.
        """
        return self._split_markdown(inpt)

    def _split_markdown(self, markdowns: list[MarkdownDataContract])-> list[MarkdownDataContract]:
        """
        Creates data rows from a batch of markdown texts by splitting them and counting tokens.
        """
        rows = []
        skipped = 0
        for s in markdowns:
            try:
                rows.extend(self.splitter.split_markdown_document(s))
            except (MarkdownException) as err:
                log.warning("skipped dokument ", extra={"reason":err.__class__.__name__,"doc":s})
                skipped += 1
        if skipped == len(markdowns):
            raise SplittException("all Documents got skipped during splitting")
        if skipped:
            log.error(f"{(skipped/len(markdowns))*100}% got skipped")
        return rows



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import warnings
from typing import Type
from ...step import TypedStep
from .data import QdrantResult
__all__=[]
try:
    _HAS_QDRANT = True
    import qdrant_client as _
except ImportError:
    _HAS_QDRANT = False
__all__=["QdrantResult", "STEPS"]
STEPS: list[Type[TypedStep]]=[]
if _HAS_QDRANT:
    from .step import QdrantConnectorStep
    from .step_mutli_vector import QdrantConnectorMultiVectorStep
    __all__.extend(["QdrantConnectorStep","QdrantConnectorMultiVectorStep"])
else:
    pass



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Type
from . import (
    milvus as __m,
    qdrant as __q,
    )
from .milvus import *
from .qdrant import *
from ..step import TypedStep
from .manual_markdown import ManualMarkdownStep
from .embedding import EmbeddingStep
__all__=[
    "TypedStep",
    "ManualMarkdownStep","EmbeddingStep"]
__all__.extend([*__q.__all__, *__m.__all__])



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from wurzel import Settings
from wurzel.datacontract import MarkdownDataContract
from wurzel.step import TypedStep


class ManualMarkdownSettings(Settings):
    """Settings fro ManMdstep"""
    FOLDER_PATH: Path

class ManualMarkdownStep(TypedStep[ManualMarkdownSettings, None,list[MarkdownDataContract]]):
    """Data Source for md files from a configurable path
    """
    def run(self, inpt: None) -> list[MarkdownDataContract]:
        return [
            MarkdownDataContract.from_file(
                fp,
                url_prefix=self.__class__.__name__ + "/"
            ) for fp in self.settings.FOLDER_PATH.rglob("*.md")
        ]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Type
from inspect import getfile
from types import NoneType
import json
from pydantic_core import PydanticUndefined
from wurzel import TypedStep, Settings
from wurzel.utils import WZ
def main(step: Type[TypedStep], gen_env = False):
    """Execute"""
    ins = WZ(step)
    set_cls: Settings = ins.settings_class
    env_prefix = step.__name__.upper()
    data = {
        'Name': step.__name__,
        'Input': 'None' if ins.input_model_class == NoneType else ins.input_model_class,
        'Output': ins.output_model_type,
        'settings': {
            'env_prefix': env_prefix,
            'fields': {k: str(v) for k,v in set_cls.model_fields.items()}
        }
    }
    if gen_env:
        setts = {True: [], False: []}
        for name, info in set_cls.model_fields.items():
            default = info.get_default(call_default_factory=True)
            default = "" if default == PydanticUndefined or default is None else default
            setts[info.is_required()].append(f"{env_prefix}__{name}={default}")
        print(f"# Env for {step.__name__} -> {getfile(step)}")
        print("# Required")
        print("\n".join(setts[True]))
        print("# Optional")
        print("\n".join(setts[False]))
    else:
        print(json.dumps(
            data,
            indent="  ",
            default=str))
